# SPDX-FileCopyrightText: 2026 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

import json
import base64

class FunctionDBValue:
    """Represents the value structure for the function database."""
    def __init__(
        self,
        lowered_name: bytes,
        ptx_data: bytes
    ):
        self.lowered_name = lowered_name
        self.ptx_data = ptx_data

    def to_bytes(self) -> bytes:
        """Serialize the value to a bytes object for storage."""
        value_dict = {
            'lowered_name': base64.b64encode(self.lowered_name).decode('utf-8'),
            'ptx_data': base64.b64encode(self.ptx_data).decode('utf-8')
        }
        return json.dumps(value_dict, sort_keys=True).encode('utf-8')

    @classmethod
    def from_bytes(cls, data: bytes) -> 'FunctionDBValue':
        """Deserialize bytes to a FunctionDBValue object."""
        value_dict = json.loads(data.decode('utf-8'))
        return cls(
            lowered_name=base64.b64decode(value_dict['lowered_name']),
            ptx_data=base64.b64decode(value_dict['ptx_data'])
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FunctionDBValue):
            return False
        return (
            self.lowered_name == other.lowered_name and
            self.ptx_data == other.ptx_data
        )

    def __repr__(self) -> str:
        return (f"FunctionDBValue(lowered_name={self.lowered_name!r}, "
                f"ptx_data={self.ptx_data!r})")


# SPDX-FileCopyrightText: 2026 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

import torch
from enum import Enum, auto
from typing import Tuple, Union

class Majorness(Enum):
    COL_MAJOR = auto()
    ROW_MAJOR = auto()
# For templates to dictate whether
# an input tensor is aligned to 16 Bytes (4 float elements)
class Alignment(Enum):
    ALIGN_1 = auto()
    ALIGN_4 = auto()

class PyTorchStreamWrapper:
    def __init__(self, pt_stream):
        self.pt_stream = pt_stream

    def __cuda_stream__(self):
        stream_id = self.pt_stream.cuda_stream
        return (0, stream_id)  # Return format required by CUDA Python

class CudaTimer:
    def __init__(self):
        """Initialize the timer, creating start and end CUDA events and recording the start time."""
        if not torch.cuda.is_available():
            raise RuntimeError("CUDA is not available")
        self.start_event = torch.cuda.Event(enable_timing=True)
        self.end_event = torch.cuda.Event(enable_timing=True)
    
    def start(self):
        """Reset the timer by recording a new start time."""
        self.start_event.record()

    def stop(self):
        """Stop the timer, record the end time, and return the elapsed time in milliseconds."""
        self.end_event.record()
        self.end_event.synchronize()  # Ensure events are complete
        return self.start_event.elapsed_time(self.end_event)

def ceil_div(x, d):
    return int((x + d - 1) // d)

def merge_batch_size(
    name : str,
    L, # previous batch size
    a : Union[float, torch.tensor],
    expected_dims : int,
    can_be_none: bool = False,
):
    if a is None:
        if not can_be_none:
            raise ValueError(f'{name} is none, but can_be_none is False')
        return L
    elif isinstance(a, float):
        if expected_dims != 0:
            raise ValueError(f'{name} is a float but expected_dims is {expected_dims}')
        return L
    elif isinstance(a, torch.Tensor):
        if a.dim() == expected_dims:
            return L
        if a.dim() == expected_dims + 1:
            this_L = a.size(0)
            if this_L != 1 and L != 1 and this_L != L:
                raise ValueError(f'{name} batch size of {this_L} does not match with another non-one batch size, {L}')
            return max(L, this_L)
        raise ValueError(f'{name} does have the right number of dimensions, expected_dims is {expected_dims} or {expected_dims+1}, got {a.dim()}')

class TensorStats:
    def __init__(self, tensor: torch.Tensor):
        if tensor.dim() not in (2, 3):
            raise ValueError("Input tensor must be 2 or 3-dimensional")
        
        if tensor.dtype != torch.float32:
            raise TypeError("Input tensor must have dtype torch.float32")
        
        dim = tensor.dim()
        shape = tensor.size()
        strides = tensor.stride()
        
        # Calculate batch information
        self._num_batches = 1 if dim == 2 else shape[0]
        self._batch_stride = 0 if dim == 2 or self._num_batches == 1 else strides[0]

        # Get row/col dimensions (last two dimensions for rank 3)
        num_rows = shape[-2]
        num_cols = shape[-1]
        stride_row = strides[-2]
        stride_col = strides[-1]

        self._shape = (self._num_batches, num_rows, num_cols)
        
        # Determine majorness
        if stride_col == 1 and stride_row >= num_cols:
            self._majorness = Majorness.ROW_MAJOR
        elif stride_row == 1 and stride_col >= num_rows:
            self._majorness = Majorness.COL_MAJOR
        else:
            raise ValueError(f"Tensor has non-standard memory layout: strides={strides}, shape={shape}")
        
        # Calculate alignment
        self._alignment_requirement_bytes = 16
        self._alignment_requirement_elements = 4
        
        self._leading_dimension_stride = max(stride_row, stride_col)
        
        is_starting_pointer_aligned = tensor.data_ptr() % self._alignment_requirement_bytes == 0
        is_leading_dimension_aligned = self._leading_dimension_stride % self._alignment_requirement_elements == 0
        
        if self._majorness == Majorness.ROW_MAJOR:
            self._alignment = Alignment.ALIGN_1
        else:
            self._alignment = Alignment.ALIGN_4 if is_starting_pointer_aligned and is_leading_dimension_aligned else Alignment.ALIGN_1

    # Accessors
    @property
    def shape(self) -> Tuple[int,int,int]:
        return self._shape

    @property
    def majorness(self) -> Majorness:
        return self._majorness
    
    @property
    def alignment(self) -> Alignment:
        return self._alignment
    
    @property
    def num_batches(self) -> int:
        return self._num_batches
    
    @property
    def batch_stride(self) -> int:
        return self._batch_stride
    
    @property
    def leading_dimension_stride(self) -> int:
        return self._leading_dimension_stride

def tensor_stats(tensor: torch.Tensor) -> TensorStats:
    return TensorStats(tensor)


# SPDX-FileCopyrightText: 2026 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

# PTX Inject now returns data type tokens as strings (e.g., "F32").
# Keep these constants for callers that want named access.
class DataTypeInfo:
    F16 = "F16"
    F16X2 = "F16X2"
    S32 = "S32"
    U32 = "U32"
    F32 = "F32"
    B32 = "B32"



# SPDX-FileCopyrightText: 2026 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

from mm_ptx.stack_ptx import (
    StackTypeEnum, 
    ArgTypeEnum, 
    create_instruction_enum,
    create_special_register_enum,
    StackPtx
)

from enum import auto, unique

@unique
class Stack(StackTypeEnum):
    f32 =   (auto(), "f32")
    s32 =   (auto(), "s32")
    u32 =   (auto(), "u32")
    pred =  (auto(), "pred")

@unique
class ArgType(ArgTypeEnum):
    f32 =   (auto(), Stack.f32,)
    u32 =   (auto(), Stack.u32,)
    pred =  (auto(), Stack.pred,)

@unique
class PtxInstruction(create_instruction_enum(ArgType)):
    copysign_f32 =          (auto(),    "copysign.f32",         [ArgType.f32, ArgType.f32],                 [ArgType.f32])
    add_ftz_f32 =           (auto(),    "add.ftz.f32",          [ArgType.f32, ArgType.f32],                 [ArgType.f32])
    sub_ftz_f32 =           (auto(),    "sub.ftz.f32",          [ArgType.f32, ArgType.f32],                 [ArgType.f32])
    mul_ftz_f32 =           (auto(),    "mul.ftz.f32",          [ArgType.f32, ArgType.f32],                 [ArgType.f32])
    fma_rn_ftz_f32 =        (auto(),    "fma.rn.ftz.f32",       [ArgType.f32, ArgType.f32, ArgType.f32],    [ArgType.f32])
    div_approx_ftz_f32 =    (auto(),    "div.approx.ftz.f32",   [ArgType.f32, ArgType.f32],                 [ArgType.f32])
    abs_ftz_f32 =           (auto(),    "abs.ftz.f32",          [ArgType.f32],                              [ArgType.f32])
    neg_ftz_f32 =           (auto(),    "neg.ftz.f32",          [ArgType.f32],                              [ArgType.f32])
    min_ftz_f32 =           (auto(),    "min.ftz.f32",          [ArgType.f32],                              [ArgType.f32])
    max_ftz_f32 =           (auto(),    "max.ftz.f32",          [ArgType.f32],                              [ArgType.f32])
    
    rcp_approx_ftz_f32 =    (auto(),    "rcp.approx.ftz.f32",   [ArgType.f32],                              [ArgType.f32])
    sqrt_approx_ftz_f32 =   (auto(),    "sqrt.approx.ftz.f32",  [ArgType.f32],                              [ArgType.f32])
    rsqrt_approx_ftz_f32 =  (auto(),    "rsqrt.approx.ftz.f32", [ArgType.f32],                              [ArgType.f32])
    sin_approx_ftz_f32 =    (auto(),    "sin.approx.ftz.f32",   [ArgType.f32],                              [ArgType.f32])
    cos_approx_ftz_f32 =    (auto(),    "cos.approx.ftz.f32",   [ArgType.f32],                              [ArgType.f32])
    lg2_approx_ftz_f32 =    (auto(),    "lg2.approx.ftz.f32",   [ArgType.f32],                              [ArgType.f32])
    ex2_approx_ftz_f32 =    (auto(),    "ex2.approx.ftz.f32",   [ArgType.f32],                              [ArgType.f32])
    tanh_approx_f32 =       (auto(),    "tanh.approx.f32",      [ArgType.f32],                              [ArgType.f32])

    setp_lt_ftz_f32 =       (auto(),    "setp.lt.ftz.f32",      [ArgType.f32, ArgType.f32],                 [ArgType.pred])
    setp_gt_ftz_f32 =       (auto(),    "setp.gt.ftz.f32",      [ArgType.f32, ArgType.f32],                 [ArgType.pred])
    selp_f32 =              (auto(),    "selp.f32",             [ArgType.f32, ArgType.f32,  ArgType.pred],  [ArgType.f32])

@unique
class SpecialRegister(create_special_register_enum(ArgType)):
    clock =     (auto(),    "clock",      ArgType.u32)
    tid_x =     (auto(),    "tid.x",      ArgType.u32)

compiler = \
    StackPtx(
        stack_enum=Stack,
        arg_enum=ArgType,
        ptx_instruction_enum=PtxInstruction, 
        special_register_enum=SpecialRegister
    )


# SPDX-FileCopyrightText: 2026 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

from .hyper_semiring import (
    HyperSemiringKernel
)

from .zoo.gemm      import Gemm
from .zoo.norm_L1   import NormL1
from .zoo.norm_L2   import NormL2
from .zoo.norm_Lp   import NormLp



# SPDX-FileCopyrightText: 2026 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

from mm_kermac import Stack, PtxInstruction
from mm_kermac.hyper_semiring_gradient import HyperSemiringGradientKernel

from enum import IntEnum
import torch

class NormLp:
    def __init__(self, epsilon):
        class Var(IntEnum):
            diff = 0,
            abs_diff = 1,
        
        self.hyper_semiring_gradient = \
            HyperSemiringGradientKernel(
                multiply_lambda=\
                    lambda reg_d, reg_b, reg_a, reg_dict: [
                        reg_b,
                        reg_d,
                        PtxInstruction.sub_ftz_f32,           # diff = [reg_d - reg_b]
                        Stack.f32.store(Var.diff),
                        Stack.load(Var.diff),                   # [diff, diff]
                        PtxInstruction.abs_ftz_f32,           # [abs(diff), diff]
                        Stack.f32.store(Var.abs_diff),

                        # if less than epsilon, clamp to epsilon
                       
                        Stack.load(Var.abs_diff),
                        Stack.f32.constant(epsilon),            # [e, abs(diff)]
                        PtxInstruction.setp_gt_ftz_f32,       # [] : [e > abs(diff)]
                        Stack.load(Var.abs_diff),
                        Stack.f32.constant(epsilon),            # [e, abs(diff)] : [e > abs(diff)]
                        PtxInstruction.selp_f32,              # [clamped_value]

                        # raise to power p_power_grad
                        PtxInstruction.lg2_approx_ftz_f32,
                        reg_dict['p_power_grad'],
                        PtxInstruction.mul_ftz_f32,
                        PtxInstruction.ex2_approx_ftz_f32,    # [pow(clamped_value, p_power_grad)]
                    
                        Stack.load(Var.diff),                   # [diff, pow(clamped_value, p_power_grad]
                        PtxInstruction.copysign_f32,          # [sign(diff) * pow(clamped_value, p_power_grad]
                        reg_a,
                        PtxInstruction.mul_ftz_f32            # [reg_a * sign(diff) * pow(clamped_value, p_power_grad]
                    ],
                accumulate_lambda=\
                    lambda reg_c, reg_diff, reg_e, _: [
                        reg_c,
                        reg_diff,
                        PtxInstruction.mul_ftz_f32,
                        reg_e,
                        PtxInstruction.add_ftz_f32
                    ],
                epilogue_lambda=\
                    lambda reg_e, _: [
                        reg_e
                    ]
            )
        
    def __call__(
        self,
        x : torch.Tensor,
        z : torch.Tensor,
        coefs : torch.Tensor,
        grad_kernel_matrix : torch.Tensor,
        p_power_grad : torch.Tensor,
        out : torch.Tensor = None,
        debug = False
    ):
        hyper_dict = {'p_power_grad': p_power_grad}
        return self.hyper_semiring_gradient(
            a = grad_kernel_matrix,
            b = x,
            c = coefs,
            d = z,
            hyper_dict=hyper_dict,
            out = out,
            debug = debug
        )



# SPDX-FileCopyrightText: 2026 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import sys
from pathlib import Path
import unittest

ROOT = Path(__file__).resolve().parents[1]
sys.path.insert(0, str(ROOT / "src"))

import torch


def _require_cuda() -> None:
    if not torch.cuda.is_available():
        raise unittest.SkipTest("CUDA is not available")
    major, _minor = torch.cuda.get_device_capability()
    if major < 8:
        raise unittest.SkipTest("Compute capability >= 8.0 required")


def _error_metrics(tensor_a: torch.Tensor, tensor_b: torch.Tensor) -> tuple[float, float, float]:
    diff = tensor_a - tensor_b
    mse = torch.mean(diff * diff)
    rmse = torch.sqrt(mse).item()
    abs_error = diff.abs()
    max_abs_error = abs_error.max().item()
    mean_abs_error = abs_error.mean().item()
    return rmse, max_abs_error, mean_abs_error


class HyperSemiringExampleTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        _require_cuda()
        torch.manual_seed(0)
        torch.cuda.manual_seed(0)

    def _assert_bounds(
        self,
        metrics: tuple[float, float, float],
        bounds: tuple[float, float, float],
        label: str,
    ) -> None:
        rmse, max_abs, mean_abs = metrics
        rmse_bound, max_bound, mean_bound = bounds
        self.assertLessEqual(rmse, rmse_bound, f"{label} rmse {rmse:.6e} > {rmse_bound:.6e}")
        self.assertLessEqual(max_abs, max_bound, f"{label} max {max_abs:.6e} > {max_bound:.6e}")
        self.assertLessEqual(mean_abs, mean_bound, f"{label} mean {mean_abs:.6e} > {mean_bound:.6e}")

    def test_hyper_semiring_example(self) -> None:
        try:
            import mm_kermac.hyper_semiring as hyper_semiring
        except Exception as exc:
            raise unittest.SkipTest(f"mm_kermac import failed: {exc}")

        device = torch.device("cuda")
        M, N, K = 128, 128, 64
        x = torch.randn(M, K, device=device)
        z = torch.randn(N, K, device=device)

        # GEMM
        out = torch.zeros(3, M, N, device=device)
        gemm = hyper_semiring.Gemm()
        kermac_out = gemm(x=x, z=z, out=out, try_to_align=False, debug=False)
        torch_out = x @ z.T
        self._assert_bounds(_error_metrics(kermac_out, torch_out), (1e-5, 1e-4, 1e-5), "Gemm")

        # L2 norm
        out = torch.zeros(3, M, N, device=device)
        norm_L2 = hyper_semiring.NormL2()
        kermac_out = norm_L2(x=x, z=z, out=out, try_to_align=False, debug=False)
        torch_out = torch.cdist(x, z, p=2.0)
        self._assert_bounds(_error_metrics(kermac_out, torch_out), (1e-4, 5e-4, 1e-4), "Norm L2")

        # L1 norm
        out = torch.zeros(3, M, N, device=device)
        norm_L1 = hyper_semiring.NormL1()
        kermac_out = norm_L1(x=x, z=z, out=out, try_to_align=False, debug=False)
        torch_out = torch.cdist(x, z, p=1.0)
        self._assert_bounds(_error_metrics(kermac_out, torch_out), (2e-3, 1e-2, 2e-3), "Norm L1")

        # Lp norms (batched)
        out = torch.zeros(3, M, N, device=device)
        p_inner = torch.tensor([1.3, 1.4, 1.5], dtype=torch.float32, device=device)
        p_outer = torch.tensor([1.0 / 1.3, 1.0 / 1.4, 1.0 / 1.5], dtype=torch.float32, device=device)
        epsilon = 1e-3

        norm_Lp = hyper_semiring.NormLp(epsilon)
        kermac_out = norm_Lp(
            x=x,
            z=z,
            p_inner=p_inner,
            p_outer=p_outer,
            out=out,
            try_to_align=False,
            debug=False,
        )

        lp_bounds = (5e-4, 2e-3, 5e-4)
        for idx, p in enumerate((1.3, 1.4, 1.5)):
            torch_out = torch.cdist(x, z, p=p)
            self._assert_bounds(_error_metrics(kermac_out[idx], torch_out), lp_bounds, f"Norm Lp p={p}")



# SPDX-FileCopyrightText: 2026 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import sys
from pathlib import Path
import unittest

ROOT = Path(__file__).resolve().parents[1]
sys.path.insert(0, str(ROOT / "src"))

import torch


def _require_cuda() -> None:
    if not torch.cuda.is_available():
        raise unittest.SkipTest("CUDA is not available")
    major, _minor = torch.cuda.get_device_capability()
    if major < 8:
        raise unittest.SkipTest("Compute capability >= 8.0 required")


def _error_metrics(tensor_a: torch.Tensor, tensor_b: torch.Tensor) -> tuple[float, float, float]:
    diff = tensor_a - tensor_b
    mse = torch.mean(diff * diff)
    rmse = torch.sqrt(mse).item()
    abs_error = diff.abs()
    max_abs_error = abs_error.max().item()
    mean_abs_error = abs_error.mean().item()
    return rmse, max_abs_error, mean_abs_error


class HyperSemiringGradientExampleTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        _require_cuda()
        torch.manual_seed(0)
        torch.cuda.manual_seed(0)

    def _assert_bounds(
        self,
        metrics: tuple[float, float, float],
        bounds: tuple[float, float, float],
        label: str,
    ) -> None:
        rmse, max_abs, mean_abs = metrics
        rmse_bound, max_bound, mean_bound = bounds
        self.assertLessEqual(rmse, rmse_bound, f"{label} rmse {rmse:.6e} > {rmse_bound:.6e}")
        self.assertLessEqual(max_abs, max_bound, f"{label} max {max_abs:.6e} > {max_bound:.6e}")
        self.assertLessEqual(mean_abs, mean_bound, f"{label} mean {mean_abs:.6e} > {mean_bound:.6e}")

    def test_hyper_semiring_gradient_example(self) -> None:
        try:
            import mm_kermac.hyper_semiring_gradient as hyper_semiring_gradient
            from mm_kermac.helpers import torch_einsum_L2_norm, torch_inefficient_p_norm
        except Exception as exc:
            raise unittest.SkipTest(f"mm_kermac import failed: {exc}")

        device = torch.device("cuda")
        M, N, O, K, L = 12, 8, 2, 32, 1

        tensor_A = torch.randn(L, K, M, device=device)  # grad_kernel_matrix
        tensor_B = torch.randn(L, N, K, device=device)  # x
        tensor_C = torch.randn(L, O, K, device=device)  # coefs
        tensor_D = torch.randn(L, N, M, device=device)  # z

        coefs = tensor_C
        grad_kernel_matrix = tensor_A
        x = tensor_B
        z = tensor_D

        epsilon = 1e-3
        bounds = (1e-4, 1e-3, 1e-4)

        # L1 grad
        norm_L1 = hyper_semiring_gradient.NormL1()
        kermac_out = norm_L1(
            x=x,
            z=z,
            coefs=coefs,
            grad_kernel_matrix=grad_kernel_matrix,
            debug=False,
        )
        torch_out = torch_inefficient_p_norm(
            x=x,
            z=z,
            coefs=coefs,
            grad_kernel_matrix=grad_kernel_matrix,
            p=1.0,
            eps=0.0,
        )
        self._assert_bounds(_error_metrics(kermac_out, torch_out), bounds, "Norm L1 Grad")

        # L2 grad
        norm_L2 = hyper_semiring_gradient.NormL2()
        kermac_out = norm_L2(
            x=x,
            z=z,
            coefs=coefs,
            grad_kernel_matrix=grad_kernel_matrix,
            debug=False,
        )
        torch_out = torch_inefficient_p_norm(
            x=x,
            z=z,
            coefs=coefs,
            grad_kernel_matrix=grad_kernel_matrix,
            p=2.0,
            eps=0.0,
        )
        einsum_out = torch_einsum_L2_norm(
            x=x,
            z=z,
            coefs=coefs,
            grad_kernel_matrix=grad_kernel_matrix,
        )
        self._assert_bounds(_error_metrics(kermac_out, torch_out), bounds, "Norm L2 Grad (inefficient)")
        self._assert_bounds(_error_metrics(kermac_out, einsum_out), bounds, "Norm L2 Grad (einsum)")

        # Lp grad
        p_power_grad = torch.tensor([1.3 - 1.0], dtype=torch.float32, device=device)
        norm_Lp = hyper_semiring_gradient.NormLp(epsilon=epsilon)
        kermac_out = norm_Lp(
            x=x,
            z=z,
            coefs=coefs,
            grad_kernel_matrix=grad_kernel_matrix,
            p_power_grad=p_power_grad,
            debug=False,
        )
        torch_out = torch_inefficient_p_norm(
            x=x,
            z=z,
            coefs=coefs,
            grad_kernel_matrix=grad_kernel_matrix,
            p=1.3,
            eps=epsilon,
        )
        self._assert_bounds(_error_metrics(kermac_out, torch_out), bounds, "Norm Lp Grad (p=1.3)")



# SPDX-FileCopyrightText: 2026 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

import argparse
import torch

from mm_kermac.helpers import compare_tensors
import mm_kermac.hyper_semiring as hyper_semiring

def parse_args():
    """Parse command-line arguments for matrix dimensions, p-norm, and flags."""
    parser = argparse.ArgumentParser(description="Run kermac.cdist_t with configurable parameters")
    parser.add_argument('-m','--M', type=int, default=10000, help='Number of rows in output matrix (default: 10000)')
    parser.add_argument('-n','--N', type=int, default=10000, help='Number of columns in output matrix (default: 10000)')
    parser.add_argument('-k','--K', type=int, default=1024, help='Inner dimension of input matrices (default: 1024)')
    parser.add_argument('-a','--try_align', default=False, action='store_true', help='Specialize kernel if tensors are 4 element aligned')
    parser.add_argument('-d','--debug', default=False, action='store_true', help='Enable debug output (default: True)')
    return parser.parse_args()

def main():
    args = parse_args()
    M, N, K = args.M, args.N, args.K
    try_to_align = args.try_align
    debug = args.debug

    device = torch.device('cuda')
    x = torch.randn(M,K,device=device)
    z = torch.randn(N,K,device=device)
    out = torch.zeros(3,M,N,device=device)

    gemm = hyper_semiring.Gemm()
    print('Running MMA')
    gemm(
        x=x, 
        z=z, 
        out=out,
        try_to_align=try_to_align,
        debug=debug
    )
    torch_out = x @ z.T
    compare_tensors(out, torch_out)

    norm_L2 = hyper_semiring.NormL2()
    print('Running Norm L2')
    norm_L2(
        x=x, 
        z=z, 
        out=out,
        try_to_align=try_to_align,
        debug=debug
    )
    torch_out = torch.cdist(x, z, p=2.0)
    compare_tensors(out, torch_out)

    norm_L1 = hyper_semiring.NormL1()
    print('Running Norm L1')
    norm_L1(
        x=x, 
        z=z, 
        out=out,
        try_to_align=try_to_align,
        debug=debug
    )
    torch_out = torch.cdist(x, z, p=1.0)
    compare_tensors(out, torch_out)

    p_inner = torch.tensor([1.3, 1.4, 1.5], dtype=torch.float32, device=device)
    p_outer = torch.tensor([1.0 / 1.3, 1.0 / 1.4, 1.0 / 1.5], dtype=torch.float32, device=device)
    epsilon = 1e-3

    norm_Lp = hyper_semiring.NormLp(epsilon)
    print('Running Norm P=1.3, P=1.4, P=1.5 Batched')
    norm_Lp(
        x=x, 
        z=z,
        p_inner=p_inner,
        p_outer=p_outer, 
        out=out,
        try_to_align=try_to_align,
        debug=debug
    )

    print('Comparing Norm P=1.3')
    torch_out = torch.cdist(x, z, p=1.3)
    compare_tensors(out[0], torch_out)

    print('Comparing Norm P=1.4')
    torch_out = torch.cdist(x, z, p=1.4)
    compare_tensors(out[1], torch_out)

    print('Comparing Norm P=1.5')
    torch_out = torch.cdist(x, z, p=1.5)
    compare_tensors(out[2], torch_out)

if __name__ == '__main__':
    main()



# SPDX-FileCopyrightText: 2026 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

from mm_kermac import PtxInstruction
from mm_kermac.hyper_semiring import HyperSemiringKernel

import torch

class NormL1:
    def __init__(self):
        self.hyper_semiring = \
            HyperSemiringKernel(
                mma_lambda=\
                    lambda reg_a, reg_b, reg_c, _: [
                        reg_a,
                        reg_b,
                        PtxInstruction.sub_ftz_f32,
                        PtxInstruction.abs_ftz_f32,
                        reg_c,
                        PtxInstruction.add_ftz_f32
                    ],
                epilogue_lambda=\
                    lambda reg_e, _: [
                        reg_e
                    ]
            )
        
    def __call__(
        self,
        x : torch.Tensor,
        z : torch.Tensor,
        out : torch.Tensor = None,
        try_to_align : bool = False,
        debug = False
    ):
        return self.hyper_semiring(
            a = x,
            b = z,
            out = out,
            hyper_dict=None,
            try_to_align = try_to_align,
            debug = debug
        )



# SPDX-FileCopyrightText: 2026 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

from mm_kermac import Stack, PtxInstruction
from mm_kermac.hyper_semiring import HyperSemiringKernel

import torch

class NormLp:
    def __init__(self, epsilon):
        self.hyper_semiring = \
            HyperSemiringKernel(
                mma_lambda=\
                    lambda reg_a, reg_b, reg_c, reg_dict: [
                        reg_a,
                        reg_b,
                        PtxInstruction.sub_ftz_f32,           # diff = reg_b - reg_a [diff]
                        PtxInstruction.abs_ftz_f32,           # [abs(diff)]

                        # raise to power p_inner
                        PtxInstruction.lg2_approx_ftz_f32,
                        reg_dict['p_inner'],
                        PtxInstruction.mul_ftz_f32,
                        PtxInstruction.ex2_approx_ftz_f32,    # [ex2(p_inner * lg2(abs(diff)))]
                                                                # [pow(abs(diff), p_inner)]
                        reg_c,
                        PtxInstruction.add_ftz_f32
                    ],
                epilogue_lambda=\
                    lambda reg_e, reg_dict: [
                        reg_e,

                        # raise to power p_outer
                        PtxInstruction.lg2_approx_ftz_f32,    #   [lg2(reg_e)]
                        reg_dict['p_outer'],                    #   [p_outer, lg2(reg_e)]
                        PtxInstruction.mul_ftz_f32,           #   [p_outer * lg2(reg_e)]
                        PtxInstruction.ex2_approx_ftz_f32,    #   [ex2(p_outer * lg2(reg_e))]
                                                                #   [pow(reg_e, p_outer)]
                        # if less than epsilon, clamp to 0
                        Stack.f32.store(0),                     # Store pow(reg_e, p_outer)
                        Stack.load(0),                          # Load it here
                        Stack.f32.constant(epsilon),            # [e, pow(reg_e, p_outer)]
                        PtxInstruction.setp_gt_ftz_f32,       # [] : [e > pow(reg_e, p_outer)]
                        Stack.load(0),                          # Load it here too
                        Stack.f32.constant(0.0),                # [0.0, pow(reg_e, p_outer)] : [e > pow(reg_e, p_outer)]
                        PtxInstruction.selp_f32               # [e > pow(reg_e, p_outer) ? 0.0 : pow(reg_e, p_outer)]
                    ]
            )
        
    def __call__(
        self,
        x : torch.Tensor,
        z : torch.Tensor,
        p_inner : torch.Tensor,
        p_outer : torch.Tensor,
        out : torch.Tensor = None,
        try_to_align : bool = False,
        debug = False
    ):
        hyper_dict = {
            'p_inner': p_inner,
            'p_outer': p_outer
        }
        return self.hyper_semiring(
            a = x,
            b = z,
            out = out,
            hyper_dict=hyper_dict,
            try_to_align = try_to_align,
            debug = debug
        )



# SPDX-FileCopyrightText: 2026 MetaMachines LLC
#
# SPDX-License-Identifier: MIT

import argparse

import torch

import mm_kermac.hyper_semiring as hyper_semiring


def _require_cuda() -> None:
    if not torch.cuda.is_available():
        raise RuntimeError("CUDA is not available")
    major, _minor = torch.cuda.get_device_capability()
    if major < 8:
        raise RuntimeError("Compute capability >= 8.0 required")


def _time_cuda(fn, iters: int) -> float:
    start = torch.cuda.Event(enable_timing=True)
    end = torch.cuda.Event(enable_timing=True)
    start.record()
    for _ in range(iters):
        fn()
    end.record()
    torch.cuda.synchronize()
    return start.elapsed_time(end) / iters


def _bench_case(name: str, kermac_fn, torch_fn, warmup: int, iters: int) -> None:
    for _ in range(warmup):
        kermac_fn()
    torch.cuda.synchronize()
    kermac_ms = _time_cuda(kermac_fn, iters)

    for _ in range(warmup):
        torch_fn()
    torch.cuda.synchronize()
    torch_ms = _time_cuda(torch_fn, iters)

    speedup = torch_ms / kermac_ms if kermac_ms > 0 else float("inf")
    print(f"{name:>8} | kermac {kermac_ms:8.3f} ms | torch {torch_ms:8.3f} ms | {speedup:6.2f}x")


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Benchmark mm-kermac hyper_semiring vs torch.cdist"
    )
    parser.add_argument("--M", type=int, default=2048, help="Rows in x (default: 2048)")
    parser.add_argument("--N", type=int, default=2048, help="Rows in z (default: 2048)")
    parser.add_argument("--K", type=int, default=256, help="Columns in x/z (default: 256)")
    parser.add_argument("--iters", type=int, default=50, help="Timed iterations (default: 50)")
    parser.add_argument("--warmup", type=int, default=10, help="Warmup iterations (default: 10)")
    parser.add_argument("--p-frac", type=float, default=1.3, help="Fractional p for NormLp (default: 1.3)")
    parser.add_argument("--epsilon", type=float, default=0.0, help="Epsilon clamp for NormLp (default: 0.0)")
    parser.add_argument("--try-align", action="store_true", help="Enable alignment specialization")
    return parser.parse_args()


def main() -> None:
    args = _parse_args()
    if args.p_frac <= 0:
        raise ValueError("--p-frac must be > 0")
    _require_cuda()

    torch.manual_seed(0)
    torch.cuda.manual_seed(0)

    device = torch.device("cuda")
    x = torch.randn(args.M, args.K, device=device)
    z = torch.randn(args.N, args.K, device=device)
    out = torch.empty((args.M, args.N), device=device)

    norm_l1 = hyper_semiring.NormL1()
    norm_l2 = hyper_semiring.NormL2()
    norm_lp = hyper_semiring.NormLp(args.epsilon)

    p_inner = torch.tensor(args.p_frac, dtype=torch.float32, device=device)
    p_outer = torch.tensor(1.0 / args.p_frac, dtype=torch.float32, device=device)

    def kermac_l1():
        norm_l1(x=x, z=z, out=out, try_to_align=args.try_align, debug=False)

    def kermac_l2():
        norm_l2(x=x, z=z, out=out, try_to_align=args.try_align, debug=False)

    def kermac_lp():
        norm_lp(
            x=x,
            z=z,
            p_inner=p_inner,
            p_outer=p_outer,
            out=out,
            try_to_align=args.try_align,
            debug=False,
        )

    def torch_l1():
        torch.cdist(x, z, p=1.0)

    def torch_l2():
        torch.cdist(x, z, p=2.0)

    def torch_lp():
        torch.cdist(x, z, p=args.p_frac)

    torch.cuda.synchronize()
    print(f"Device: {torch.cuda.get_device_name(device)}")
    print(f"M={args.M} N={args.N} K={args.K} iters={args.iters} warmup={args.warmup}")
    print(f"Fractional p={args.p_frac} epsilon={args.epsilon} try_align={args.try_align}")
    print("   case |    kermac ms |     torch ms | speedup")

    with torch.no_grad():
        _bench_case("p=1.0", kermac_l1, torch_l1, args.warmup, args.iters)
        _bench_case("p=2.0", kermac_l2, torch_l2, args.warmup, args.iters)
        _bench_case(f"p={args.p_frac:g}", kermac_lp, torch_lp, args.warmup, args.iters)


if __name__ == "__main__":
    main()
